#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>
#include <glib-object.h>

#include <courierauth.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <auth/auth.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>

#include "authvtoaster.h"

static void log_handler(const gchar *domain, GLogLevelFlags level,
                        const gchar *message, gpointer user_data)
{
	if ((level & G_LOG_FLAG_FATAL) && !(level & G_LOG_FLAG_RECURSION)) {
		g_setenv("CHECKVPW_ERROR", message, TRUE);
		v_exec_script("checkvpw-error");
	}
}

int auth_vtoaster_pre(const char *user, const char *service,
                      int (*cbfunc)(struct authinfo *auth, void *data),
                      void *data)
{
	return authvtoastercommon(user, 0, cbfunc, data);
}

static int do_auth_vtoaster(const char *user, struct authinfo *auth);

static VDomainEntry *entry = NULL;
static VDomain *domain = NULL;
static VUser *user = NULL;
static gchar *username = NULL, *userdomain = NULL;

int authvtoastercommon(const char *user, const char *pass,
                       int (*cbfunc)(struct authinfo *auth, void *data),
                       void *data)
 {
        VDomainAuthClass *klass;
	VDomainAuth *auth = NULL;
	struct authinfo authinfo;
	int retval;

	memset(&authinfo, 0, sizeof authinfo);

	retval = do_auth_vtoaster(user, &authinfo);
	if (retval)
		return retval;

	if (time(NULL) >= entry->expiry)
		return -1;

	auth = g_object_new(V_TYPE_AUTH, NULL);
	if (!auth)
	    	return -1;

	klass = V_DOMAIN_AUTH_GET_CLASS(auth);
	klass->setup(auth, entry);
	retval = klass->auth(auth, pass, NULL);
	if (auth)
		g_object_unref(auth);

	if (!retval)
		return -1;

	authinfo.clearpasswd = pass;
	return ((*cbfunc)(&authinfo, data));
}

static int do_auth_vtoaster(const char *login, struct authinfo *authinfo)
{
	static gboolean initialized = FALSE;
	if (!initialized) {
		static char *args[] = { "authvtoaster", NULL };
		char **argv = args;
		int argc = 1;
		v_init(&argc, &argv, NULL, NULL, NULL, "");
		g_log_set_default_handler(log_handler, NULL);
		initialized = TRUE;
	} else {
		g_free(username);
		g_free(userdomain);
		if (entry)
			g_object_unref(entry);
		if (user)
			g_object_unref(user);
		if (domain)
			g_object_unref(domain);

		username = NULL;
		userdomain = NULL;
		entry = NULL;
		user = NULL;
		domain = NULL;
	}

	v_split_login_name(login, &username, &userdomain);
	if (!userdomain) {
		userdomain = g_strdup(v_get_hostname());
		if (!userdomain)
			return -1;
	}

	if (!username || !*username)
		return -1;

	if (!*userdomain)
		return -1;

	domain = v_domain_new(NULL, userdomain, FALSE, NULL);
	if (!domain) 
		return -1;
	v_domain_set_auth_method(domain, V_TYPE_AUTH);

	user = v_domain_get_user(domain);

	entry = v_domain_get_entry(domain, username, NULL);
	if (!entry)
		return -1;

	authinfo->sysusername = user->user;
	authinfo->sysuserid = &user->uid;
	authinfo->sysgroupid = user->gid;
	authinfo->homedir = user->home;
	authinfo->address = login;
	authinfo->maildir = entry->maildir;
	authinfo->fullname = entry->name;
	authinfo->passwd = entry->passwd;

	return 0;
}

void authvtoasterclose(void)
{
	g_free(username);
	g_free(userdomain);
	if (entry)
		g_object_unref(entry);
	if (user)
		g_object_unref(user);
	if (domain)
		g_object_unref(domain);

	username = NULL;
	userdomain = NULL;
	entry = NULL;
	user = NULL;
	domain = NULL;
}
