#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <errno.h>

#include <glib.h>
#include <gmodule.h>

#include <misc/user.h>

struct _VUserClass {
	GObjectClass		parent_class;
};

G_DEFINE_TYPE(VUser, v_user, G_TYPE_OBJECT);

static void v_user_finalize(GObject *gobject);

static void v_user_class_init(VUserClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = v_user_finalize;
}

static void v_user_init(VUser *user)
{
}

static void v_user_finalize(GObject *gobject)
{
	VUser *user = V_USER(gobject);

	g_free(user->user);
	user->user = NULL;

	g_free(user->home);
	user->home = NULL;

	g_free(user->shell);
	user->shell = NULL;

	G_OBJECT_CLASS(v_user_parent_class)->finalize(gobject);
}

VUser *v_user_new(const gchar *name)
{
	VUser *user = NULL;

	if (name) {
		struct passwd pwbuf, *pw = NULL;
		gsize buflen = sysconf(_SC_GETPW_R_SIZE_MAX);
		void *buf;

		buf = g_malloc(buflen);
		if (getpwnam_r(name, &pwbuf, buf, buflen, &pw) == 0) {
			user = g_object_new(V_TYPE_USER, NULL);
			user->user = g_strdup(pw->pw_name);
			user->home = g_strdup(pw->pw_dir);
			user->shell = g_strdup(pw->pw_shell);
			user->uid = pw->pw_uid;
			user->gid = pw->pw_gid;
		}
		memset(&pwbuf, 0, sizeof pwbuf);
		memset(buf, 0, buflen);
		g_free(buf);
	} else {
		user = g_object_new(V_TYPE_USER, NULL);
		user->user = g_strdup(g_getenv("USER"));
		user->home = g_strdup(g_getenv("HOME"));
		user->uid = getuid();
		user->gid = getgid();
	}

	return user;
}

gboolean v_user_become(VUser *user)
{
	GError *error = NULL;
	gchar *msg = NULL;

	g_return_val_if_fail(V_IS_USER(user), FALSE);

	if (getuid() != user->uid)  {
		if (initgroups(user->user, user->gid) < 0) {
			msg = g_strdup_printf(_("Error setting supplementary "
			                        "groups for user `%s': %%s"),
			                      user->user);
			goto out_errno;
		}

		if (!v_drop_process_rights(user->user, user->uid, user->gid,
		                           &error))
			goto out;
	}

	if (chdir(user->home) < 0) {
		msg = g_strdup_printf(_("Error changing to home "
		                        "directory `%s': %%s"), user->home);
		goto out_errno;
	}

	if (!g_setenv("USER", user->user, TRUE) ||
	    !g_setenv("LOGNAME", user->user, TRUE) ||
	    !g_setenv("HOME", user->home, TRUE) ||
	    !g_setenv("SHELL", user->shell, TRUE)) {
		msg = g_strdup(_("Error setting environment variables: %s"));
		goto out_errno;
	}

	return TRUE;

out_errno:
	g_set_error(&error, G_FILE_ERROR,
	            g_file_error_from_errno(errno),
	            msg, g_strerror(errno));
out:
	if (error) {
		g_warning(msg, error->message);
		g_error_free(error);
	}
	g_free(msg);
	return FALSE;
}
