#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <glib.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <qmail/qmail.h>

#define BUFFER_SIZE		4096
#define QMAIL_DEFAULT_ROOT	"/var/qmail"
#define QMAIL_QUEUE_PROGRAM	"/bin/qmail-queue"
#define CHECKPW_PROTOCOL_FD 	3
#define CHECKPW_PROTOCOL_LEN	512

G_CONST_RETURN gchar *v_qmail_get_root(void)
{
	static gchar *qmail_root = NULL;

	if (qmail_root)
		return qmail_root;

	qmail_root = v_config_get_string(NULL, NULL, "qmail-root");
	if (!qmail_root)
		qmail_root = QMAIL_DEFAULT_ROOT;

	return qmail_root;
}

gchar **v_qmail_get_headers(VMaildirHeaders mask, GError **error)
{
	static const gchar const *headers[] = { "UFLINE", "RPLINE", "DTLINE" };
	gchar **result;
	int i, j;

	g_return_val_if_fail(!error || !*error, NULL);

	for(i = j = 0; i < G_N_ELEMENTS(headers); i++)
		if (mask & (1 << i))
			j++;

	result = g_new0(gchar *, j + 1);

	for(i = j = 0; i < G_N_ELEMENTS(headers); i++) {
		const gchar *header = headers[i];
		const gchar *value = g_getenv(header);

		if (!value) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_INVALID,
			            _("Env variable `%s' is not set\n"),
			            header);
			goto out;
		}

		if (mask & (1 << i))
			result[j++] = g_strdup(value);
	}

	return result;

out:
	g_strfreev(result);
	return NULL;
}

static void v_qmail_inject_helper(gpointer user_data)
{
	int *fds = user_data;

	close(fds[1]);
	close(fds[3]);
	if (dup2(fds[0], 0) != 0 || dup2(fds[2], 1) != 1)
		exit(81);
}

struct inject_info {
	GMainLoop	*loop;
	gint		result;
};

static void v_qmail_inject_watch(GPid pid, gint status, gpointer data)
{
	struct inject_info *info = data;

	info->result = status;
	g_main_loop_quit(info->loop);
}

gboolean v_qmail_inject(GIOChannel *in, const gchar *sender,
                        gchar **recipients, GError **error)
{
	struct inject_info info = { NULL, 0 };
	GIOChannel *io[2] = { NULL, NULL };
	GIOStatus status;
	gboolean  result = FALSE;
	gsize size, count;
	gchar **argv = NULL;
	gchar *buf = NULL, *tmp;
	sig_t oldsig = SIG_DFL;
	GPid pid;
	int fds[4] = { -1, -1, -1, -1 };
	int setup = 0, i;

	g_return_val_if_fail(in, -1);
	g_return_val_if_fail(sender, -1);
	g_return_val_if_fail(recipients, -1);
	g_return_val_if_fail(!error || !*error, -1);

	if (g_io_channel_get_encoding(in)) {
		status = g_io_channel_set_encoding(in, NULL, error);
		if (status != G_IO_STATUS_NORMAL)
			return -1;
	}

	argv = g_new0(gchar *, 2);
	argv[0] = g_strconcat(v_qmail_get_root(), QMAIL_QUEUE_PROGRAM, NULL);
	argv[1] = NULL;

	if (pipe(&fds[0]) < 0 || pipe(&fds[2]) < 0) {
		g_set_error(error, G_FILE_ERROR,
		            g_file_error_from_errno(errno),
		            g_strerror(errno));
		if (fds[0] >= 0) {
			close(fds[0]);
			close(fds[1]);
		}
		goto out;
	}

	oldsig = signal(SIGPIPE, SIG_IGN);
	setup = 1;

	result = g_spawn_async(NULL, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD,
	                       v_qmail_inject_helper, fds, &pid, error);
	if (!result)
		goto out;

	for(i = 0; i < 2; i++) {
		close(fds[i * 2]);
		fds[i * 2] = -1;
		io[i] = g_io_channel_unix_new(fds[i * 2 + 1]);
		if (!io[i]) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_FAILED,
			            "Unable to create new descriptor for "
			            "fd %d.", i);
			goto out;
		}
		fds[i * 2 + 1] = -1;
		g_io_channel_set_close_on_unref(io[i], TRUE);
		status = g_io_channel_set_encoding(io[i], NULL, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
	}

	setup = 2;
	result = FALSE;

	buf = g_malloc(BUFFER_SIZE);
	for(;;) {
		status = g_io_channel_read_chars(in, buf, BUFFER_SIZE,
		                                 &size, error);
		if (status != G_IO_STATUS_NORMAL &&
		    status != G_IO_STATUS_EOF)
			goto out;

		if (!size)
			break;

		status = g_io_channel_write_chars(io[0], buf, size,
		                                  &count, error);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		if (count != size) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_SHORT_WRITE,
			            "Short write while queueing mail.");
			goto out;
		}
	}
	g_free(buf);
	buf = NULL;

	status = g_io_channel_flush(io[0], error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	g_io_channel_unref(io[0]);
	io[0] = NULL;

	tmp = g_strconcat("F", sender, NULL);
	size = strlen(tmp) + 1;
	status = g_io_channel_write_chars(io[1], tmp, size, &count, error);
	g_free(tmp);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != size) {
		g_set_error(error, V_QMAIL_ERROR,
		            V_QMAIL_ERROR_SHORT_WRITE,
		            "Short write while queueing mail.");
		goto out;
	}

	for(; *recipients; recipients++) {
		tmp = g_strconcat("T", *recipients, NULL);
		size = strlen(tmp) + 1;
		status = g_io_channel_write_chars(io[1], tmp, size,
		                                  &count, error);
		g_free(tmp);
		if (status != G_IO_STATUS_NORMAL)
			goto out;
		if (count != size) {
			g_set_error(error, V_QMAIL_ERROR,
			            V_QMAIL_ERROR_SHORT_WRITE,
			            "Short write while queueing mail.");
			goto out;
		}
	}

	status = g_io_channel_write_chars(io[1], "", 1, &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count != 1) {
		g_set_error(error, V_QMAIL_ERROR,
		            V_QMAIL_ERROR_SHORT_WRITE,
		            "Short write while queueing mail.");
		goto out;
	}

	status = g_io_channel_flush(io[1], error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	g_io_channel_unref(io[1]);
	io[1] = NULL;

	info.loop = g_main_loop_new(NULL, TRUE);
	g_child_watch_add(pid, v_qmail_inject_watch, &info);
	g_main_loop_run(info.loop);
	g_main_loop_unref(info.loop);

	result = info.result == 0;
out:
	if (!result) {
		if (info.result)
			g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_FAILED,
			            _("qmail-queue exited with error %d."),
			            info.result);
		else
			g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_FAILED,
			            _("Spawning qmail-queue failed."));
	}

	if (setup >= 2)
		g_spawn_close_pid(pid);
	if (setup >= 1)
		signal(SIGPIPE, oldsig);

	for(i = 0; i < 2; i++)
		if (io[i])
			g_io_channel_unref(io[i]);
	for(i = 0; i < 4; i++)
		close(fds[i]);

	g_free(buf);
	g_strfreev(argv);

	return result;
}

gboolean v_qmail_checkpassword_protocol(gchar **username, gchar **password,
                                        GError **error)
{
	struct stat st;
	GIOChannel *io = NULL;
	GIOStatus status;
	gboolean retval = FALSE;
	gsize count;
	gchar *p, *buf = NULL;

	g_return_val_if_fail(username && !*username, FALSE);
	g_return_val_if_fail(password && !*password, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	if (fstat(CHECKPW_PROTOCOL_FD, &st) < 0) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Could not open protocol fd");
		goto out;
	}

	io = g_io_channel_unix_new(CHECKPW_PROTOCOL_FD);
	if (!io) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Could not open protocol fd");
		goto out;
	}
	g_io_channel_set_close_on_unref(io, TRUE);
	g_io_channel_set_buffered(io, FALSE);
	status = g_io_channel_set_encoding(io, NULL, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	status = g_io_channel_set_flags(io, 0, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	buf = g_malloc(CHECKPW_PROTOCOL_LEN);
	status = g_io_channel_read_chars(io, buf, CHECKPW_PROTOCOL_LEN,
	                                 &count, error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;
	if (count >= CHECKPW_PROTOCOL_LEN) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Read buffer too long");
		goto out;
	}
	buf[count] = 0;

	p = buf;
	count = strlen(p);
	if (p + count >= buf + CHECKPW_PROTOCOL_LEN) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Invalid data received");
		goto out;
	}

	*username = g_strdup(p);

	p += count + 1;
	count = strlen(p);
	if (p + count >= buf + CHECKPW_PROTOCOL_LEN) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            "Invalid data received");
		g_free(*username);
		*username = NULL;
		goto out;
	}
	*password = g_strdup(p);

	retval = TRUE;
out:
	if (buf) {
		memset(buf, 0, CHECKPW_PROTOCOL_LEN);
		g_free(buf);
	}
	if (io)
		g_io_channel_unref(io);

	return retval;
}

gboolean v_qmail_get_environment(gchar **_user, gchar **_ext, gchar **_host,
                                 gchar **_sender, GError **error) {
	const gchar *user, *ext, *host, *sender, *msg = NULL;

	g_return_val_if_fail(!_user || !*_user, FALSE);
	g_return_val_if_fail(!_ext || !*_ext, FALSE);
	g_return_val_if_fail(!_host || !*_host, FALSE);
	g_return_val_if_fail(!_sender || !*_sender, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	user = g_getenv("USER");
	ext = g_getenv("EXT");
	host = g_getenv("HOST");
	sender = g_getenv("SENDER");

	if (!user)
		msg = "USER";
	else if (!ext)
		msg = "EXT";
	else if (!host)
		msg = "HOST";
	else if (!sender)
		msg = "SENDER";
		
	if (msg) {
		g_set_error(error, V_QMAIL_ERROR, V_QMAIL_ERROR_INVALID,
		            _("Env variable %s is not set"), msg);
		return FALSE;
	}

	if (_user)
		*_user = g_strdup(user);
	if (_ext)
		*_ext = g_strdup(ext);
	if (_host)
		*_host = g_strdup(host);
	if (_sender)
		*_sender = g_strdup(sender);

	return TRUE;
}
