#include <stdarg.h>
#include <time.h>

#include <glib.h>
#include <gmodule.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <vdomain/vdomain.h>
#include "internal.h"

struct _VDomainEntryClass {
	GObjectClass	parent_class;
};

G_DEFINE_TYPE(VDomainEntry, v_domain_entry, G_TYPE_OBJECT);

static void v_domain_entry_finalize(GObject *gobject);

static void v_domain_entry_class_init(VDomainEntryClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = v_domain_entry_finalize;
}

static void v_domain_entry_init(VDomainEntry *entry)
{
	entry->hardquota = G_MAXUINT;
	entry->softquota = G_MAXUINT;
	entry->msgsize = G_MAXUINT;
	entry->msgcount = G_MAXUINT;
	entry->ctime = time(NULL);
	entry->expiry = TIME_T_MAX;
	entry->enabled = TRUE;
}

static void v_domain_entry_finalize(GObject *gobject)
{
	VDomainEntry *entry = V_DOMAIN_ENTRY(gobject);

	g_free(entry->name);
	entry->name = NULL;

	g_free(entry->maildir);
	entry->maildir = NULL;

	g_strfreev(entry->forwards);
	entry->forwards = NULL;

	g_free(entry->personal);
	entry->personal = NULL;

	G_OBJECT_CLASS(v_domain_entry_parent_class)->finalize(gobject);
}

VDomainEntry *v_domain_entry_new(const gchar *name, const gchar *maildir)
{
	VDomainEntry *entry;

	g_return_val_if_fail(name, NULL);

	entry = g_object_new(V_TYPE_DOMAIN_ENTRY, NULL);
	entry->name = g_strstrip(g_ascii_strdown(name, -1));
	entry->maildir = g_strdup(maildir);

	return entry;
}

VDomainEntry *v_domain_entry_copy(const VDomainEntry *old)
{
	VDomainEntry *entry;

	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(old), NULL);

	entry = g_object_new(V_TYPE_DOMAIN_ENTRY, NULL);
	entry->name = g_strdup(old->name);
	entry->maildir = g_strdup(old->maildir);
	entry->passwd = g_strdup(old->passwd);
	entry->forwards = g_strdupv(old->forwards);
	entry->personal = g_strdup(old->personal);

	entry->hardquota = old->hardquota;
	entry->softquota = old->softquota;
	entry->msgsize = old->msgsize;
	entry->msgcount = old->msgcount;
	entry->ctime = old->ctime;
	entry->expiry = old->expiry;

	entry->enabled = old->enabled;

	return entry;
}

void v_domain_entry_load_defaults(VDomainEntry *entry)
{
	gulong lval;
	guint val;

	if (v_config_get_uint(NULL, "defaults", "hardquota", &val) &&
	    val < G_MAXUINT)
		entry->hardquota = val;

	if (v_config_get_uint(NULL, "defaults", "softquota", &val) &&
	    val < G_MAXUINT)
		entry->softquota = val;

	if (v_config_get_uint(NULL, "defaults", "msgsize", &val) &&
	    val < G_MAXUINT)
		entry->msgsize = val;

	if (v_config_get_uint(NULL, "defaults", "msgcount", &val) &&
	    val < G_MAXUINT)
		entry->msgcount = val;

	if (v_config_get_ulong(NULL, "defaults", "expiry", &lval)) {
		if (lval < TIME_T_MAX &&
		    (time_t)(entry->ctime + lval) >= entry->ctime)
			entry->expiry = entry->ctime + lval;
		else
			entry->expiry = TIME_T_MAX;
	}
}

void v_domain_entry_set_name(VDomainEntry *entry, const gchar *name)
{
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_free(entry->name);
	entry->name = g_strdup(name);
}

void v_domain_entry_set_maildir(VDomainEntry *entry, const gchar *maildir)
{
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_free(entry->maildir);
	entry->maildir = g_strdup(maildir);
}

void v_domain_entry_set_password(VDomainEntry *entry, const gchar *pass)
{
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_free(entry->passwd);
	entry->passwd = (pass && pass[0]) ? g_strdup(pass) : NULL;
}

void v_domain_entry_set_forwards_v(VDomainEntry *entry, gchar **forwards)
{
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_strfreev(entry->forwards);
	entry->forwards = (forwards && forwards[0]) ?
	                  		g_strdupv(forwards) : NULL;
}

void v_domain_entry_set_forwards(VDomainEntry *entry, ...)
{
	GPtrArray *array = g_ptr_array_new();
	va_list args;
	gchar **tmp;
	gchar *arg;
	guint len;

	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_strfreev(entry->forwards);

	va_start(args, entry);
	while((arg = va_arg(args, gchar *)) != NULL)
		g_ptr_array_add(array, g_strdup(arg));
	g_ptr_array_add(array, NULL);
	va_end(args);

	len = array->len;
	tmp = (gchar **)g_ptr_array_free(array, FALSE);
	entry->forwards = g_renew(gchar *, tmp, len);
}

void v_domain_entry_set_personal(VDomainEntry *entry, const gchar *personal)
{
	g_return_if_fail(V_IS_DOMAIN_ENTRY(entry));

	g_free(entry->personal);
	entry->personal = (personal && personal[0]) ? g_strdup(personal)
	                                            : NULL;
}

gchar *v_domain_entry_dump(VDomainEntry *entry)
{
	gchar *result, *tmp;

	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(entry), NULL);

	tmp = entry->forwards ? g_strjoinv(" ", entry->forwards) : NULL;
	result = g_strjoin(" ", entry->name,
	                   entry->maildir ? entry->maildir : "-", tmp, NULL);
	g_free(tmp);

	return result;
}

gchar *v_domain_entry_output(VDomainEntry *entry)
{
	GString *result = NULL;

	g_return_val_if_fail(V_IS_DOMAIN_ENTRY(entry), NULL);

	result = g_string_new("");

	if (entry->forwards) {
		gchar *tmp = g_strjoinv(" ", entry->forwards);
		g_string_append_printf(result, "FORWARDS='%s'\n", tmp);
		g_free(tmp);
	}

	if (entry->maildir)
		g_string_append_printf(result, "MAILDIR='%s'\n", entry->maildir);

	g_string_append_printf(result, "VUSER='%s'\n", entry->name);
	if (entry->personal)
		g_string_append_printf(result, "VUSER_PERSONAL='%s'\n", entry->personal);

	g_string_append_printf(result, "VUSER_CTIME='%lu'\n", (gulong)entry->ctime);
	g_string_append_printf(result, "VUSER_EXPIRY='%lu'\n", (gulong)entry->expiry);
	g_string_append_printf(result, "VUSER_MSGCOUNT='%u'\n", entry->msgcount);
	g_string_append_printf(result, "VUSER_MSGSIZE='%u'\n", entry->msgsize);
	g_string_append_printf(result, "VUSER_HARDQUOTA='%u'\n", entry->hardquota);
	g_string_append_printf(result, "VUSER_SOFTQUOTA='%u'\n", entry->softquota);
	g_string_append_printf(result, "VUSER_MAILBOX_ENABLED='%c'\n", entry->enabled ? '1' : '0');

	return g_string_free(result, FALSE);
}
