#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <misc/vconfig.h>
#include <auth/auth.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>

#define	DEFAULT_MAILDIR_ARG_STR	"maildir"

enum {
	CHECKVPW_OK,
	CHECKVPW_FAIL_LOGIN,
	CHECKVPW_FAIL_TEMP,
	CHECKVPW_FAIL_ERROR
};

static void log_handler(const gchar *domain, GLogLevelFlags level,
                        const gchar *message, gpointer user_data)
{
	if ((level & G_LOG_FLAG_FATAL) && !(level & G_LOG_FLAG_RECURSION)) {
		g_setenv("CHECKVPW_ERROR",message, TRUE);
		v_exec_script("checkvpw-error");
		exit(2);
	}
}

static void replace_maildir_arg(int argc, char **argv, const gchar *maildir)
{
	gchar *arg = v_config_get_string(NULL, NULL, "maildir-arg-str");
	int i;

	if (!arg)
		arg = DEFAULT_MAILDIR_ARG_STR;
	for(i = 1; i < argc; i++)
		if (strcmp(argv[i], arg) == 0)
			break;
	if (arg != DEFAULT_MAILDIR_ARG_STR)
		g_free(arg);

	if (i < argc)
		argv[i] = g_strdup(maildir);

	g_setenv("MAILDIR", maildir, TRUE);
}

static gboolean auth(const gchar *login, const gchar *password,
                     gchar **path, GError **error)
{
	gboolean retval = FALSE;
	VAccount *account = NULL;
	VMaildir *maildir = NULL;
	VDomain *domain = NULL;
	VUser *user = NULL;
	gchar *username = NULL, *userdomain = NULL;

	g_return_val_if_fail(login, FALSE);
	g_return_val_if_fail(password, FALSE);
	g_return_val_if_fail(path && !*path, FALSE);
	g_return_val_if_fail(!error || !*error, FALSE);

	v_split_login_name(login, &username, &userdomain);
	if (!userdomain) {
		userdomain = g_strdup(v_get_hostname());
		if (!userdomain) {
			g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_FAILED,
			            "Error determining local host name");
			goto out;
		}
	}

	if (!username || !*username) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
		            "Empty user name");
		goto out;
	}

	if (!*userdomain) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
		            "Empty domain name");
		goto out;
	}

	domain = v_domain_new(NULL, userdomain, FALSE, NULL);
	if (!domain) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_NOENT,
		            "Invalid or unknown base user or domain");
		goto out;
	}
	v_domain_set_auth_method(domain, V_TYPE_AUTH);

	if (v_exec_script("checkvpw-presetuid") != 0) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_FAILED,
		            "Failure to exec checkvpw-presetuid");
		goto out;
	}

	user = v_domain_get_user(domain);
	if (!v_user_become(user)) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_FAILED,
		            "Failure to drop user rights");
		goto out;
	}

	account = v_domain_auth_account(domain, username, password, error);
	if (!account)
		goto out;

	maildir = v_account_get_maildir(account);
	if (!maildir || !(*path = v_maildir_get_base(maildir))) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_INVALID,
		            "User maildir invalid");
		goto out;
	}

	if (v_exec_script("checkvpw-postsetuid") != 0) {
		g_set_error(error, V_AUTH_ERROR, V_AUTH_ERROR_FAILED,
		            "Failure to exec checkvpw-postsetuid");
		goto out;
	}

	retval = TRUE;
out:
	g_free(username);
	g_free(userdomain);
	if (maildir)
		g_object_unref(maildir);
	if (account)
		g_object_unref(account);
	if (user)
		g_object_unref(user);
	if (domain)
		g_object_unref(domain);

	return retval;
}

int main(int argc, char **argv)
{
	GError *error = NULL;
	gchar *login = NULL, *password = NULL;
	gchar *path = NULL;
	int retcode = 111;
	int retval = CHECKVPW_FAIL_ERROR;

	v_init(&argc, &argv, NULL, NULL, NULL, "");

	g_log_set_default_handler(log_handler, NULL);

	if (!v_qmail_checkpassword_protocol(&login, &password, &error))
		goto out;

	retval = CHECKVPW_FAIL_LOGIN;

	if (!auth(login, password, &path, &error)) {
		if (error && error->domain == V_AUTH_ERROR &&
		    error->code == V_AUTH_ERROR_FAILED)
			retval = CHECKVPW_FAIL_TEMP;
		goto out;
	}

	replace_maildir_arg(argc, argv, path);

	retval = CHECKVPW_OK;
out:
	if (password) {
		memset(password, 0, strlen(password));
		g_free(password);
	}
	g_free(login);
	g_free(path);

	if (retval == CHECKVPW_OK && argc >= 2) {
		retcode = v_exec(argv + 1, &error);

		if (retcode) {
			if (!error)
				g_set_error(&error, G_SPAWN_ERROR,
				            G_SPAWN_ERROR_FAILED,
				            "Execution of command line "
				            "arguments failed");
			retval = CHECKVPW_FAIL_TEMP;
		}
	}

	if (error) {
		gchar *tmp = g_locale_from_utf8(error->message, -1, NULL,
		                                NULL, NULL);

		g_setenv("CHECKVPW_ERROR", tmp, TRUE);
		g_free(tmp);
		g_error_free(error);
	}

	switch(retval) {
	case CHECKVPW_OK:
		return 0;
	case CHECKVPW_FAIL_LOGIN:
		g_usleep(v_config_get_int(NULL, NULL,
		                          "password-delay") * 1000000);
		v_exec_script("checkvpw-loginfail");
		return 1;
	case CHECKVPW_FAIL_TEMP:
		v_exec_script("checkvpw-error");
		return retcode;
	case CHECKVPW_FAIL_ERROR:
	default:
		v_exec_script("checkvpw-error");
		return 2;
	}
}
