#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <auth/auth.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>

static VDomainEntry *entry = NULL;
static gchar **aliases = NULL;
static gboolean quiet = FALSE;
static gboolean no_password = FALSE;
static gboolean no_maildir = FALSE;

static inline gboolean parse_guint(const gchar *value, guint *retval)
{
	gulong val;
	gchar *end = NULL;

	val = strtoul(value, &end, 10);
	if (end && *end)
		return FALSE;
	if (val > G_MAXUINT)
		return FALSE;

	*retval = (guint)val;

	return TRUE;
}

static inline void ensure_entry(void)
{
	if (!entry) {
		entry = v_domain_entry_new("", NULL);
		v_domain_entry_load_defaults(entry);
	}
}

#define SET_ERROR(e)	G_STMT_START{					\
	g_set_error(error, G_OPTION_ERROR,				\
	            G_OPTION_ERROR_BAD_VALUE, "%s in option `%s'",	\
	            (e), option_name);					\
	return FALSE;	}G_STMT_END

#define SET_ERROR_IF_FAIL(c, e)						\
	G_STMT_START{ if (!(c)) SET_ERROR(e); }G_STMT_END

#define CODE_CB_TEMPLATE(name, code)					\
	static gboolean name##_cb(const gchar *option_name,		\
	                          const gchar *value,			\
	                          gpointer data, GError **error)	\
	{								\
		ensure_entry();						\
		{ code }						\
		return TRUE;						\
	}

#define GUINT_CB_TEMPLATE(name)						\
	CODE_CB_TEMPLATE(name,						\
		SET_ERROR_IF_FAIL(parse_guint(value, &entry->name),	\
		                  _("Bad value"));)

#define SET_METHOD_CB_TEMPLATE(name)					\
	CODE_CB_TEMPLATE(name, v_domain_entry_set_##name(entry, value);)

GUINT_CB_TEMPLATE(msgcount)
GUINT_CB_TEMPLATE(msgsize)
GUINT_CB_TEMPLATE(hardquota)
GUINT_CB_TEMPLATE(softquota)

SET_METHOD_CB_TEMPLATE(maildir);
SET_METHOD_CB_TEMPLATE(personal);

CODE_CB_TEMPLATE(expiry, {
	gulong val;
	gchar *end = NULL;

	val = strtoul(value, &end, 10);
	SET_ERROR_IF_FAIL(!end || !*end, _("Bad value"));

	ensure_entry();

	if (val <= TIME_T_MAX &&
	    (time_t)(entry->ctime + val) >= entry->ctime)
		entry->expiry = entry->ctime + val;
	else
		entry->expiry = TIME_T_MAX;
})

static const GOptionEntry entries[] = {
	{ "msgcount",		'c', 0, G_OPTION_ARG_CALLBACK,		msgcount_cb,	"Set the user's message count limit",		"INT"	},
	{ "no-mailbox",		'D', 0, G_OPTION_ARG_NONE,		&no_maildir,	"Do not create a mailbox for this user",	NULL	},
	{ "directory",		'd', 0, G_OPTION_ARG_CALLBACK,		maildir_cb,	"Set the path to the user's mailbox",		"VALUE"	},
	{ "expiry",		'e', 0, G_OPTION_ARG_CALLBACK,		expiry_cb,	"Set the account's expiry time (in seconds)",	"INT"	},
	{ "forward",		'f', 0, G_OPTION_ARG_STRING_ARRAY,	&aliases,	"Add a forwarding address to this user",	"ITEM"	},
	{ "no-password",	'P', 0, G_OPTION_ARG_NONE,		&no_password,	"Do not ask for a password",			NULL	},
	{ "personal",		'p', 0, G_OPTION_ARG_CALLBACK,		personal_cb,	"Set the user's personal information",		"VALUE"	},
	{ "hardquota",		'Q', 0, G_OPTION_ARG_CALLBACK,		hardquota_cb,	"Set the user's hard quota (in bytes)",		"INT"	},
	{ "softquota",		'q', 0, G_OPTION_ARG_CALLBACK,		softquota_cb,	"Set the user's soft quota (in bytes)",		"INT"	},
	{ "quiet",		0,   0, G_OPTION_ARG_NONE,		&quiet,		"Suppress all status messages",			NULL	},
	{ "msgsize",		'z', 0, G_OPTION_ARG_CALLBACK,		msgsize_cb,	"Set the user's message size limit (in bytes)",	"INT"	},
	{ NULL }
};

int main(int argc, char **argv)
{
	VDomainEntry *alias_entry = NULL;
	VMaildir *maildir = NULL;
	VDomain *domain;
	GError *error = NULL;
	gchar **alias;
	gchar *tmp;
	gint retval = 1;

	v_init(&argc, &argv, entries, NULL, _("USERNAME [ALIAS...]"),
	       _("Add a user to a virtual domain"));

	if (argc < 2)
		v_init_error(_("Too few command-line arguments"));

	domain = v_domain_new(NULL, NULL, TRUE, NULL);
	if (!domain) {
		fprintf(stderr, _("%s: Could not build default domain\n"),
		        g_get_prgname());
		return 1;
	}
	v_domain_set_auth_method(domain, V_TYPE_AUTH);

	ensure_entry();

	tmp = g_strstrip(g_ascii_strdown(argv[1], -1));
	v_domain_entry_set_name(entry, tmp);
	g_free(tmp);

	if (!no_password) {
		tmp = v_get_password();
		if (!tmp)
			goto out;
		v_domain_set_entry_password(domain, entry, tmp);
		g_free(tmp);
	}

	if (!entry->maildir && !no_maildir) {
		tmp = v_domain_build_userdir(domain, entry->name);
		v_domain_entry_set_maildir(entry, tmp);
		g_free(tmp);
	}

	v_domain_entry_set_forwards_v(entry, argv + 2);

	if (!v_domain_add_entry(domain, entry, &error)) {
		if (!quiet)
			fprintf(stderr, _("%s: error adding the virtual "
			                  "user `%s': %s\n"), g_get_prgname(),
			        entry->name, error->message);
		goto out;
	}

	if (entry->maildir &&
	    !(maildir = v_maildir_new_create(entry->maildir, NULL)))
		goto out;

	for(alias = aliases; alias && *alias; alias++) {
		alias_entry = v_domain_entry_copy(entry);
		v_domain_entry_set_name(alias_entry, *alias);
		v_domain_entry_set_password(alias_entry, NULL);
		v_domain_entry_set_maildir(alias_entry, NULL);
		v_domain_entry_set_forwards(alias_entry, entry->name, NULL);

		if (!v_domain_add_entry(domain, alias_entry, &error)) {
			if (!quiet)
				fprintf(stderr, _("%s: error adding the "
				                  "virtual alias `%s': %s\n"),
					g_get_prgname(), alias_entry->name,
					error->message);
			goto out;
		}

		g_object_unref(alias_entry);
		alias_entry = NULL;
	}

	if (!v_domain_commit_changes(domain, NULL)) {
		if (!quiet)
			fprintf(stderr, _("%s: Error commiting changes to "
			                  "the database.\n"), g_get_prgname());
		goto out;
	}

	if (!quiet)
		printf(_("%s: user `%s' successfully added.\n"),
		       g_get_prgname(), entry->name);

	retval = 0;
out:
	if (error)
		g_error_free(error);
	if (alias_entry)
		g_object_unref(alias_entry);
	if (entry)
		g_object_unref(entry);
	if (domain)
		g_object_unref(domain);
	return retval;
}
