#include <stdio.h>

#include <glib.h>
#include <glib-object.h>

#include <misc/utils.h>
#include <qmail/qmail.h>
#include <vdomain/vdomain.h>

enum {
	VDELIVER_OK,
	VDELIVER_TEMP,
	VDELIVER_FAIL
};

static gboolean quiet = FALSE;
static gboolean delivered = TRUE;
static gboolean from = FALSE;
static gboolean retpath = TRUE;
static gboolean no_delivered = FALSE;
static gboolean no_from = FALSE;
static gboolean no_retpath = FALSE;

static const GOptionEntry entries[] = {
	{ NULL,			'D', 0, G_OPTION_ARG_NONE,		&delivered,	"Add a \"Delivered-To:\" line (default)",	NULL	},
	{ NULL,			'F', 0, G_OPTION_ARG_NONE,		&from,		"Add a \"From \" mailboxline",			NULL	},
	{ NULL,			'R', 0, G_OPTION_ARG_NONE,		&retpath,	"Add a \"Return-Path:\" line (default)",	NULL	},
	{ NULL,			'd', 0, G_OPTION_ARG_NONE,		&no_delivered,	"Do not add a \"Delivered-To:\" line",		NULL	},
	{ NULL,			'f', 0, G_OPTION_ARG_NONE,		&no_from,	"Do not add a \"From \" mailboxline (default)",	NULL	},
	{ "quiet",		0,   0, G_OPTION_ARG_NONE,		&quiet,		"Suppress all status messages",			NULL	},
	{ NULL,			'r', 0, G_OPTION_ARG_NONE,		&no_retpath,	"Do not add a \"Return-Path:\" line",		NULL	},
	{ NULL }
};

int main(int argc, char **argv)
{
	GIOChannel *mail = NULL;
	VAccount *account = NULL;
	VDomain *domain = NULL;
	GError *error = NULL;
	gchar **headers = NULL;
	gchar *user = NULL, *host = NULL, *sender = NULL;
//	gchar *msg = NULL;
	gchar *tmp;
	gint retval = VDELIVER_FAIL;

	v_init(&argc, &argv, entries, NULL, NULL,
	       _("VToaster delivery agent"));
	if (no_delivered)
		delivered = FALSE;
	if (no_from)
		from = FALSE;
	if (no_retpath)
		retpath = FALSE;

	if (!v_qmail_get_environment(NULL, &user, &host, &sender, NULL)) {
//		if (!quiet)
//			printf("%s: %s\n", g_get_prgname(), msg);
		goto out;
	}

	tmp = g_strstrip(g_ascii_strdown(user, -1));
	g_free(user);
	user = tmp;

	headers = v_qmail_get_headers((from      ? V_MAILDIR_UFLINE : 0) |
	                              (retpath   ? V_MAILDIR_RPLINE : 0) |
	                              (delivered ? V_MAILDIR_DTLINE : 0), NULL);

	domain = v_domain_new(NULL, NULL, FALSE, NULL);
	if (!domain) {
		if (!quiet)
			printf(_("%s: Could not initialize domain handler.\n"),
			       g_get_prgname());
		goto out;
	}

	account = v_domain_get_account(domain, user, NULL);
	if (!account) {
		if (!quiet)
			printf(_("%s: Invalid or unknown virtual user `%s'\n"),
			       g_get_prgname(), user);
		goto out;
	}

	g_object_unref(domain);
	domain = NULL;

	mail = g_io_channel_unix_new(0);
	g_io_channel_set_encoding(mail, NULL, NULL);

	v_account_set_environent(account);

	if (v_exec_script("vdeliver-predeliver") != 0) {
		if (!quiet)
			printf(_("%s: Execution of vdeliver-predeliver "
			         "failed."), g_get_prgname());
		goto out;
	}

	if (!v_account_deliver(account, mail, sender, host, headers, NULL)) {
		if (!quiet)
			printf(_("%s: Execution of vdeliver-predeliver "
			         "failed."), g_get_prgname());
		goto out;
	}

	if (v_exec_script("vdeliver-postdeliver") != 0) {
		if (!quiet)
			printf(_("%s: Execution of vdeliver-postdeliver "
			         "failed."), g_get_prgname());
		goto out;
	}

	retval = VDELIVER_OK;
out:
	if (error)
		g_error_free(error);
	g_free(user);
	g_free(host);
	g_free(sender);
	if (mail)
		g_io_channel_unref(mail);
	if (account)
		g_object_unref(account);
	if (domain)
		g_object_unref(domain);

	switch(retval) {
		case VDELIVER_OK:
			return 0;
		case VDELIVER_FAIL:
			return 100;
		case VDELIVER_TEMP:
		default:
			return 111;
	}
}
